/* This program simulates several different load balancing  */
/* strategies for Zoom servers (e.g., RANDOM, ROUND_ROBIN). */
/* It reads a 7-column input file of Zoom meeting info, as  */
/* generated from makeworkload.c when the MEETINGS_ONLY and */
/* DEBUG options are both enabled. The output format is     */
/* time series data of load balancing metrics for plotting. */
/*                                                          */
/* Usage: cc -o balance balance.c                           */
/*        ./balance < meetings.dat > stats.dat              */
/*                                                          */
/* Written by Carey Williamson              June 14, 2022   */

#include <stdio.h>
#include <stdlib.h>
#include <time.h> /* for optional seeding of PRNG */

#define NUM_MEETINGS 4000
#define MAX_MEETINGS 100
#define MAX_USERS 8000
#define INFINITY 9999999.9

#define NUM_SERVERS 10

#define BASERATE 1.0  /* Mbps for video stream */

// Report the meeting and/or user occupancy distribution at the end
//#define MTGOCCUPANCY 1
//#define USEROCCUPANCY 1

/* Policies for scheduling and load balancing */
//#define SINGLE_SERVER 1
#define RANDOM 1
//#define ROUND_ROBIN 1
//#define FEWEST_MEETINGS 1
//#define FEWEST_USERS 1
//#define LEAST_WORK 1
//#define BANDWIDTH 1
//#define MOST_MEETINGS 1
//#define MOST_USERS 1
//#define SIZEMOD10 1  /* choose servers based on meeting size % 10 */
//#define TIME_BASED 1   /* choose servers based on time of day */
#define TIME_STEP 120
//#define PARTITION 1   /* use high-numbered servers for large meetings */
#define FRACTION 0.8  /* fraction of servers for small meetings */
#define THRESHOLD 10  /* threshold size limit for small meetings */

/* Verbose debugging, including start/end events and input reading */
//#define DEBUG 1

int main()
  {

    float starts[NUM_MEETINGS], ends[NUM_MEETINGS], duration[NUM_MEETINGS];
    int participants[NUM_MEETINGS], servers[NUM_MEETINGS];
    int meetings[NUM_SERVERS], users[NUM_SERVERS];
    int mymtgs[NUM_SERVERS], myusers[NUM_SERVERS];
    float work[NUM_SERVERS], mywork[NUM_SERVERS], totwork;  /* in user-minutes */
    float bw[NUM_SERVERS], mybw[NUM_SERVERS], totbw;  /* in Mbps */
    int lowindex, highindex;
#ifdef MTGOCCUPANCY
    int mtgoccupancy[MAX_MEETINGS+1];
#endif
#ifdef USEROCCUPANCY
    int useroccupancy[MAX_USERS+1];
#endif
    int minmtg, maxmtg, minusers, maxusers;
    int totmtg, totusers, idlers;
    float timestamp, dur, nextstart, nextend, leastwork, mostwork, minbw, maxbw;
    int i, j, num, conns, server, startindex, endindex;
    int meetingID, connID, count, done;
    char src, dst;
    
    num = 0;
    while ( scanf("%f %d %d %d %c %c %f", &timestamp, &meetingID, &connID, &count, &src, &dst, &dur) != EOF)
      { 
	starts[num] = timestamp;
	ends[num] = timestamp + dur;
	duration[num] = dur;
	participants[num] = count;
	servers[num] = -1;
#ifdef DEBUG
	printf("Meeting %d with %d users duration %3.1f from %f to %f\n",
	       meetingID, count, duration[num], starts[num], ends[num]);
#endif
	num++; 
      }
    starts[num] = INFINITY;
    ends[num] = INFINITY;
    
    fprintf(stderr, "Read in %d lines of input\n", num);

#ifdef SINGLE_SERVER
    printf("SINGLE SERVER\n");
#endif
#ifdef RANDOM
    printf("RANDOM\n");
#endif
#ifdef ROUND_ROBIN
    printf("ROUND ROBIN\n");
#endif
#ifdef FEWEST_MEETINGS
    printf("FEWEST MEETINGS\n");
#endif
#ifdef MOST_MEETINGS
    printf("MOST MEETINGS\n");
#endif
#ifdef FEWEST_USERS
    printf("FEWEST USERS\n");
#endif
#ifdef MOST_USERS
    printf("MOST USERS\n");
#endif
#ifdef SIZEMOD10
    printf("SIZE MOD 10\n");
#endif
#ifdef TIME_BASED
    printf("TIME_BASED %d\n", TIME_STEP);
#endif
#ifdef PARTITION
    printf("PARTITION %d\n", THRESHOLD);
#endif
    /* initialize servers to empty */
    for( i = 0; i < NUM_SERVERS; i++ )
      {
	meetings[i] = 0;
	users[i] = 0;
	work[i] = 0.0;
	bw[i] = 0.0;
	mymtgs[i] = 0;
	myusers[i] = 0;
	mywork[i] = 0.0;
	mybw[i] = 0.0;
      }
#ifdef MTGOCCUPANCY
    for( i = 0; i <= MAX_MEETINGS; i++ )
      {
	mtgoccupancy[i] = 0;
      }
#endif
#ifdef USEROCCUPANCY
    for( i = 0; i <= MAX_USERS; i++ )
      {
	useroccupancy[i] = 0;
      }
#endif
    done = 0;
    totmtg = 0;
    totusers = 0;
    totwork = 0.0;
    totbw = 0.0;
    lowindex = 0;
    highindex = (int) FRACTION * NUM_SERVERS;
    /* use a fixed seed for repeatable randomness */
    srandom(12345);
    /* use a time-varying seed for complete randomness */
    srandom(time(NULL));
    
loopit:
    
    /* find the next meeting to start */
    startindex = -1;
    nextstart = INFINITY;
    for( i = 0; i < num; i++ )
      {
	if( starts[i] < nextstart )
	  {
	    nextstart = starts[i];
	    startindex = i;
	  }
      }
#ifdef DEBUG
    printf("Next meeting to start is %d with %d users at time %f\n",
	    startindex, participants[startindex], starts[startindex]);
#endif

    /* find the next meeting to end */
    endindex = -1;
    nextend = INFINITY;
    for( i = 0; i < num; i++ )
      {
	if( ends[i] < nextend )
	  {
	    nextend = ends[i];
	    endindex = i;
	  }
      }
#ifdef DEBUG
    printf("Next meeting to end is %d with %d users at time %f\n",
	    endindex, participants[endindex], ends[endindex]);
#endif

    if( nextstart < nextend )
      {
	timestamp = nextstart;
	starts[startindex] = INFINITY;

	/* choose a server */
#ifdef SINGLE_SERVER
	server = 0;
#endif
#ifdef RANDOM
	server = random() % NUM_SERVERS;
#endif
#ifdef ROUND_ROBIN
	server = startindex % NUM_SERVERS;
#endif
#ifdef SIZEMOD10
	server = participants[startindex] % 10;
#endif
#ifdef TIME_BASED
	server = ((int) timestamp / TIME_STEP) % NUM_SERVERS;
#endif
#ifdef PARTITION
	if( participants[startindex] <= THRESHOLD )
	  {
	    /* assign small meetings to low-numbered servers */
	    server = lowindex;
	    lowindex++;
	    if( lowindex == FRACTION * NUM_SERVERS )
	      lowindex = 0;
	  }
	else
	  {
	    /* assign large meetings to high-numbered servers */
	    server = highindex;
	    highindex++;
	    if( highindex == NUM_SERVERS )
	      highindex = FRACTION * NUM_SERVERS;
	  }
#endif

	/* check meeting and user counts */
	minmtg = MAX_MEETINGS;
	maxmtg = -1;
	minusers = MAX_USERS;
	maxusers = -1;
	leastwork = MAX_MEETINGS*MAX_USERS;
	mostwork = -1.0;
	minbw = MAX_USERS*BASERATE;
	maxbw = -1.0*BASERATE;
	idlers = 0;
	for( i = 0; i < NUM_SERVERS; i++ )
	  {
#ifdef MTGOCCUPANCY
	    mtgoccupancy[meetings[i]]++;
#endif
#ifdef USEROCCUPANCY
	    useroccupancy[users[i]]++;
#endif
	    if( meetings[i] < minmtg )
	      {
		minmtg = meetings[i];
#ifdef FEWEST_MEETINGS
		server = i;
#endif
	      }
	    if( meetings[i] > maxmtg )
	      {
		maxmtg = meetings[i];
#ifdef MOST_MEETINGS
		server = i;
#endif
	      }
	    if( users[i] < minusers )
	      {
		minusers = users[i];
#ifdef FEWEST_USERS
		server = i;
#endif
	      }
	    if( users[i] > maxusers )
	      {
		maxusers = users[i];
#ifdef MOST_USERS
		server = i;
#endif
	      }
	    if( work[i] < leastwork )
	      {
		leastwork = work[i];
#ifdef LEAST_WORK
		server = i;
#endif
	      }
	    if( work[i] > mostwork )
	      {
		mostwork = work[i];
	      }
	    if( bw[i] < minbw )
	      {
		minbw = bw[i];
#ifdef BANDWIDTH
		server = i;
#endif
	      }
	    if( bw[i] > maxbw )
	      {
		maxbw = bw[i];
	      }
	    if( meetings[i] == 0 )
	      idlers++;
	  }

	/* report stats just before adding this meeting */
	printf("Time- %f minmtg %d maxmtg %d minuser %d maxuser %d totmtg %d totuser %d minbw %3.1f maxbw %3.1f totbw %3.1f leastwork %3.1f mostwork %3.1f totwork %3.1f idle %d\n",
	       timestamp, minmtg, maxmtg, minusers, maxusers, totmtg, totusers, minbw, maxbw, totbw, leastwork, mostwork, totwork, idlers);

	/* assign to the chosen server */
	servers[startindex] = server;
	
	/* update server load */
	mymtgs[server]++;
	myusers[server] += participants[startindex];
	meetings[server]++;
	if( meetings[server] == MAX_MEETINGS )
	  fprintf(stderr, "Server %d reached MAX_MEETINGS %d at time %f\n",
		  server, MAX_MEETINGS, timestamp);;
	users[server] += participants[startindex];
	if( users[server] >= MAX_USERS )
	  fprintf(stderr, "Server %d reached %d users (MAX_USERS %d) at time %f\n",
		  server, users[server], MAX_USERS, timestamp);;
	work[server] += participants[startindex]*duration[startindex]/60.0;
	mywork[server] += participants[startindex]*duration[startindex]/60.0;
	if( participants[startindex] > 2 )
	  {
	    bw[server] += participants[startindex]*BASERATE;
	    mybw[server] += participants[startindex]*BASERATE;
	    totbw += participants[startindex]*BASERATE;
	  }
	totmtg++;
	totusers += participants[startindex];
	totwork += participants[startindex]*duration[startindex]/60.0;
#ifdef DEBUG
	printf("Added meeting %d with %d users dur %3.1f work %3.1f bw %3.1f to server %d at time %f\n",
	       startindex, participants[startindex], duration[startindex],
	       participants[startindex]*duration[startindex],
	       (participants[startindex] > 2 ? participants[startindex]*BASERATE : 0.0),
	       server, timestamp);
	printf("Server %d now has %d meetings %d users %3.1f work %3.1f bw\n",
	       server, meetings[server], users[server], work[server], bw[server]);
#endif
	/* report stats just after adding this meeting */
	minmtg = MAX_MEETINGS;
	maxmtg = -1;
	minusers = MAX_USERS;
	maxusers = -1;
	leastwork = MAX_MEETINGS*MAX_USERS;
	mostwork = -1.0;
	minbw = MAX_USERS*BASERATE;
	maxbw = -1.0*BASERATE;
	idlers = 0;
	for( i = 0; i < NUM_SERVERS; i++ )
	  {
	    if( meetings[i] < minmtg )
	      minmtg = meetings[i];
	    if( meetings[i] > maxmtg )
	      maxmtg = meetings[i];
	    if( users[i] < minusers )
	      minusers = users[i];
	    if( users[i] > maxusers )
	      maxusers = users[i];
	    if( work[i] < leastwork )
	      leastwork = work[i];
	    if( work[i] > mostwork )
	      mostwork = work[i];
	    if( bw[i] < minbw )
	      minbw = bw[i];
	    if( bw[i] > maxbw )
	      maxbw = bw[i];
	    if( meetings[i] == 0 )
	      idlers++;
	  }
	printf("Time+ %f minmtg %d maxmtg %d minuser %d maxuser %d totmtg %d totuser %d minbw %3.1f maxbw %3.1f totbw %3.1f leastwork %3.1f mostwork %3.1f totwork %3.1f idle %d\n",
	       timestamp, minmtg, maxmtg, minusers, maxusers, totmtg, totusers, minbw, maxbw, totbw, leastwork, mostwork, totwork, idlers);
      }
    else
      {
	timestamp = nextend;
	ends[endindex] = INFINITY;

	/* report stats just before removing this meeting */
	minmtg = MAX_MEETINGS;
	maxmtg = -1;
	minusers = MAX_USERS;
	maxusers = -1;
	leastwork = MAX_MEETINGS*MAX_USERS;
	mostwork = -1.0;
	minbw = MAX_USERS*BASERATE;
	maxbw = -1.0*BASERATE;
	idlers = 0;
	for( i = 0; i < NUM_SERVERS; i++ )
	  {
#ifdef MTGOCCUPANCY
	    mtgoccupancy[meetings[i]]++;
#endif
#ifdef USEROCCUPANCY
	    useroccupancy[users[i]]++;
#endif
	    if( meetings[i] < minmtg )
	      minmtg = meetings[i];
	    if( meetings[i] > maxmtg )
	      maxmtg = meetings[i];
	    if( users[i] < minusers )
	      minusers = users[i];
	    if( users[i] > maxusers )
	      maxusers = users[i];
	    if( work[i] < leastwork )
	      leastwork = work[i];
	    if( work[i] > mostwork )
	      mostwork = work[i];
	    if( bw[i] < minbw )
	      minbw = bw[i];
	    if( bw[i] > maxbw )
	      maxbw = bw[i];
	    if( meetings[i] == 0 )
	      idlers++;
	  }
	printf("time- %f minmtg %d maxmtg %d minuser %d maxuser %d totmtg %d totuser %d minbw %3.1f maxbw %3.1f totbw %3.1f leastwork %3.1f mostwork %3.1f totwork %3.1f idle %d\n",
	       timestamp, minmtg, maxmtg, minusers, maxusers, totmtg, totusers, minbw, maxbw, totbw, leastwork, mostwork, totwork, idlers);

	/* find the server hosting this meeting */
	server = servers[endindex];
	
	/* update server load */
	meetings[server]--;
	users[server] -= participants[endindex];
	if( meetings[server] == 0 )
	  work[server] = 0.0;
	else work[server] -= participants[endindex]*duration[endindex]/60.0;
	if( participants[endindex] > 2 )
	  {
	    bw[server] -= participants[endindex]*BASERATE;
	    totbw -= participants[endindex]*BASERATE;
	  }
	totmtg--;
	totusers -= participants[endindex];
	if( totmtg == 0 )
	  totwork = 0.0;
	else totwork -= participants[endindex]*duration[endindex]/60.0;
#ifdef DEBUG
	printf("Removed meeting %d with %d users dur %3.1f work %3.1f bw %3.1f from server %d at time %f\n",
	       endindex, participants[endindex], duration[endindex],
	       participants[endindex]*duration[endindex],
	       (participants[endindex] > 2 ? participants[endindex]*BASERATE : 0.0),
	       server, timestamp);
	printf("Server %d now has %d meetings %d users %3.1f work %3.1f bw\n",
	       server, meetings[server], users[server], work[server], bw[server]);
#endif

	done++;

	/* check meeting and user counts */
	minmtg = MAX_MEETINGS;
	maxmtg = -1;
	minusers = MAX_USERS;
	maxusers = -1;
	leastwork = MAX_MEETINGS*MAX_USERS;
	mostwork = -1.0;
	minbw = MAX_USERS*BASERATE;
	maxbw = -1.0*BASERATE;
	idlers = 0;
	for( i = 0; i < NUM_SERVERS; i++ )
	  {
	    if( meetings[i] < minmtg )
	      minmtg = meetings[i];
	    if( meetings[i] > maxmtg )
	      maxmtg = meetings[i];
	    if( users[i] < minusers )
	      minusers = users[i];
	    if( users[i] > maxusers )
	      maxusers = users[i];
	    if( work[i] < leastwork )
	      leastwork = work[i];
	    if( work[i] > mostwork )
	      mostwork = work[i];
	    if( bw[i] < minbw )
	      minbw = bw[i];
	    if( bw[i] > maxbw )
	      maxbw = bw[i];
	    if( meetings[i] == 0 )
	      idlers++;
	  }
	/* report stats just after removing this meeting */
	printf("time+ %f minmtg %d maxmtg %d minuser %d maxuser %d totmtg %d totuser %d minbw %3.1f maxbw %3.1f totbw %3.1f leastwork %3.1f mostwork %3.1f totwork %3.1f idle %d\n",
	       timestamp, minmtg, maxmtg, minusers, maxusers, totmtg, totusers, minbw, maxbw, totbw, leastwork, mostwork, totwork, idlers);
      }

    if( done < num )
      goto loopit;

    /* report stats on meetings hosted by chosen servers */
    printf("mymtgs[i]:");
    for(i = 0; i < NUM_SERVERS; i++ )
      printf(" %d %d", i, mymtgs[i]);
    printf("\n");

    /* report stats on users hosted by chosen servers */
    printf("myusers[i]:");
    for(i = 0; i < NUM_SERVERS; i++ )
      printf(" %d %d", i, myusers[i]);
    printf("\n");

    /* report stats on work done by chosen servers */
    printf("mywork[i]:");
    for(i = 0; i < NUM_SERVERS; i++ )
      printf(" %d %3.1f", i, mywork[i]);
    printf("\n");

    /* report stats on bandwidth consumption by chosen servers */
    printf("mybw[i]:");
    for(i = 0; i < NUM_SERVERS; i++ )
      printf(" %d %3.1f", i, mybw[i]);
    printf("\n");

#ifdef MTGOCCUPANCY
    /* report stats on meeting occupancy distribution */
    printf("Server occupancy distribution (in meetings):\n");
    int subtotal = 0;
    for(i = 0; i <= MAX_MEETINGS; i++ )
      {
	subtotal += mtgoccupancy[i];
	printf("%d %d %8.6f %8.6f\n",
	       i, mtgoccupancy[i], mtgoccupancy[i]/(num*2.0*NUM_SERVERS), subtotal/(num*2.0*NUM_SERVERS));
      }
#endif

#ifdef USEROCCUPANCY
    /* report stats on user occupancy distribution */
    printf("Server occupancy distribution (in users):\n");
    for(i = 0; i <= MAX_USERS; i++ )
      printf("%d %d\n", i, useroccupancy[i]);
#endif

    //    printf("The End!\n");
  }
