/* This program generates synthetic Zoom call workloads.    */
/*                                                          */
/* Usage: cc -o makeworkload makeworkload.c -lm             */
/*        ./makeworkload                                    */
/*                                                          */
/* Written by Carey Williamson              June 12, 2022   */
/* Revised and improved                 February 17, 2023   */

/* Header files */
#include <stdio.h>
#include <stdlib.h>
#include <math.h> /* use "man 3 log" for the math functions */
#include <time.h> /* for optional seeding of PRNG */

extern float powf(float,float);
extern double pow(double,double);

/* Manifest constants */
#define NUM_VALUES 3600 /* number of Zoom meetings to generate */
#define MEAN_IAT 60.0  /* seconds */
#define SHORT_IAT 2.0  /* seconds */
#define CLASS_DURATION 3600.0  /* seconds */
#define MIN_DURATION 60.0  /* seconds */
#define MAX_DURATION 21600.0  /* seconds */
#define MEAN_DURATION 1800.0  /* seconds */
#define MEAN_SIZE 10.0  /* participants */
#define P2P_PROB 0.4  /* P2P Zoom */
#define SHORT_PROB 0.5 /* short P2P meeting */
#define SMALL_PROB 0.83 /* at most 10 users */
#define ZIPF_MAX LIMIT_LARGE /* truncate tail of Zipf distribution */
#define CLASS_PROB 0.6 /* Zoom class */
#define HOME_PROB 0.5  /* at home (not on campus) */
#define WIFI_PROB 0.8  /* WiFi instead of Ethernet */
#define VANCOUVER_PROB 0.5  /* Zoom server in Vancouver */
#define TORONTO_PROB 0.4  /* Zoom server in Toronto */

/* New debugging option to explicitly identify meetings vs sessions */
#define MEETINGS_VIEW 1

/* Use these settings to get small meetings with 1-10 users */
#define LIMIT_SMALL 10
#define THETA_SMALL 2.27

/* Use these settings for larger meetings with 11-320 users */
#define LIMIT_LARGE 300
#define THETA_LARGE 1.62

// These next two options are older ways of doing things.
// They do still work, but the inherent problem is that the PRNG
// sample path you get from MEETINGS_ONLY will be different from
// what you get when you generate user sessions. The MEETINGS_VIEW
// option above is preferable to get all the info at the same time,
// which you can post-process to extract what you are looking for.
// If you do choose to activate these, the please disable MEETINGS_VIEW
// so that you do not get redundant debugging output.

// Default setting is for user session level generation (called "conns").
// Note that sessions sometimes start prior to the meeting host starting,
// so you likely need to sort your workload file into chronological order.
// If you want just the meetings level, then uncomment the next line.
//#define MEETINGS_ONLY 1

/* Full debugging option for each user session of each meeting, */
/* useful when doing detailed statistical analysis of sessions. */
/* Comment this line out to make workload files for the routing */
/* simulation, but uncomment it for load balancing simulation. */
//#define DEBUG 1

/***********************************************************************/
/*                 RANDOM NUMBER GENERATION STUFF                      */
/***********************************************************************/

/* Parameters for random number generation. */
#define MAX_INT 2147483647       /* Maximum positive integer 2^31 - 1 */

/* array used for initializing seeds for Normal distribution */
unsigned short Nvar1_seed[3] = {4,7,2};
unsigned short Nvar2_seed[3] = {5,8,1};
unsigned short Nvar3_seed[3] = {9,6,3};

/* Generate a random floating point number uniformly distributed in [0,1] */
float Uniform01()
  {
    float randnum;
    /* get a random positive integer from random() */
    randnum = (float) 1.0 * random();
    /* divide by max int to get something in between 0.0 and 1.0 */
    randnum = (float) randnum / (1.0 * MAX_INT); 
    return( randnum );
  }


/* Generate a random floating point number from an exponential    */
/* distribution with mean mu.                                     */
float Exponential(mu)
    float mu;
 {
    float randnum, ans;

    /* The algorithm here is simple. Start with a U(0,1) random */
    /* number, and then inverse map it through the cumulative   */
    /* distribution function (CDF) for the exponential distn.   */
    randnum = Uniform01();
    ans = -(mu) * log(randnum);
    return( ans );
  }


/* Generate random positive integer geometrically distributed with mean m */
int Geometric(m)
    double m;
  {
    double p;    /* p is the probability of "success" for each trial */
    int k;

    k = 1;
    p = 1.0 / m;                          /* the inverse of the mean */
    while( Uniform01() > p )
      k++;
    return( k );
  }


/* the rejection method is used for normal distribution 
   (a) Generate two uniform U(0,1) variates u1 and u2. 
   (b) Let x = -log(u1).
   (c) If u2 > e to the power -(x - 1)square by 2, go back to step
   [a].
   (d) Generate u3.
   (e) If u3 > 0.5, return (mean + (std * x)); otherwise return
   (mean - (std * x)). */

double Normal(float mean, float std)
{
  double var_u1, var_u2, var_u3;
  double x, tmp_val;
  
  var_u1 = erand48(Nvar1_seed);
  var_u2 = erand48(Nvar2_seed);
  x = -log(var_u1);
  tmp_val = exp(-((x - 1)*(x - 1)) / 2);
  while(var_u2 > tmp_val)
    {
      var_u1 = erand48(Nvar1_seed);
      var_u2 = erand48(Nvar2_seed);
      x = -log(var_u1);
      tmp_val = exp(-((x - 1)*(x - 1)) / 2);
    }
  var_u3 = erand48(Nvar3_seed);
  if(var_u3 > 0.5)
    return(mean + (std * x));
  else 
    return(mean - (std * x));
}


/* Generate values from a Zipf-like distribution with given slope */
int Zipf(slope, limit)
     float slope;
     int limit;
  {
    int i, j;
    float prob[ZIPF_MAX+1], cumprob[ZIPF_MAX+1];
    float sumprob, p;

    /* Initialize */
    sumprob = 0.0;
    cumprob[0] = 0.0;
    
    /* Compute raw values for each i */
    for( i = 1; i <= limit; i++ )
      {
	prob[i] = 1.0 / powf((float)i, (float)slope);
#ifdef ZIPF_DEBUG
	printf("Prob[%d] is %8.6f\n", i, prob[i]);
#endif
	sumprob += prob[i];
      }
#ifdef ZIPF_DEBUG
    printf("Sum is %8.6f\n", sumprob);
#endif

    /* Normalize these to get probabilities */
    for( i = 1; i <= limit; i++ )
      {
	prob[i] /= sumprob;
	cumprob[i] = prob[i] + cumprob[i-1];
#ifdef ZIPF_DEBUG
	printf("Normalized Prob[%d] is %8.6f    Cumulative: %8.6f\n",
	       i, prob[i], cumprob[i]);
#endif
      }

    /* Now make a value from this distribution */
    p = Uniform01();
    i = 1;
    while( p > cumprob[i] )
      i++;
    
    return(i);
  }

/***********************************************************************/
/*                 MAIN PROGRAM                                        */
/***********************************************************************/
int main(int argc, char *argv[])
  {
    int i, j, conns, size, flag;
    float uni, iat, dur, timer, starttime, endtime;
    char source, dest;

    /* Set initial seed for random number generator.        */
    //    srandom(12753);        /* fixed seed for initial testing */
    //    srandom(54791); /* decent example 1 */
    srandom(33567); /* decent example 2 */  /* for routing sim example */
    //    srandom(95731); /* decent example 3 */
    //    srandom(25948); /* decent example 4 */
    //    srandom(17591); /* decent example 5 */
    //    srandom(13570); /* decent example 6 */
// Enable the line below for time-of-day seeding for different workload each run */
    //    srandom(time(NULL));   /* based on time of day */

    timer = 0.0;
    conns = 0;
    
    for(i = 0; i < NUM_VALUES; i++)
      {
	/* Determine the number of participants for the meeting */
	uni = Uniform01();
	if( uni < P2P_PROB )
	  size = 2;
#ifdef GEOMETRIC
	/* Use a shifted Geometric distribution for non-P2P meetings */
	else size = Geometric(MEAN_SIZE) + 2;
#else
	/* Use a Zipf-like distribution for all meetings */
	else
	  {
	    uni = Uniform01();
	    if( uni < SMALL_PROB )
	      {
		size = Zipf(THETA_SMALL, LIMIT_SMALL);
		/* reject values for N=1 and N=2 though! */
		while( size <= 2 )
		  size = Zipf(THETA_SMALL, LIMIT_SMALL);
	      }
	    else
	      {
		size = Zipf(THETA_LARGE, LIMIT_LARGE);
		/* reject values that are too small */
		while( size <= LIMIT_SMALL )
		  size = Zipf(THETA_LARGE, LIMIT_LARGE);
	      }
	  }
#endif

	if( size == 2 )  /* peer-to-peer Zoom */
	  {
	    /* Generate duration from a hybrid distribution */
	    uni = Uniform01();
	    if( uni < SHORT_PROB )
	      {
		/* short P2P meeting */
		dur = MEAN_DURATION + Normal(0, 900.0);
		while( dur < 0.0 )
		  dur = MEAN_DURATION + Normal(0, 900.0);
		/* add slight bias (longer durations) for wee hours of morning (before 8am) */
		if( timer < 27000.0 )
		    dur += Uniform01()*CLASS_DURATION + Exponential(MEAN_DURATION);
	      }
	    else
	      {
		dur = Exponential(1.75*MEAN_DURATION);
		/* add slight bias (longer durations) for wee hours of morning (before 8am) */
		if( timer < 27000.0 )
		    dur += Uniform01()*CLASS_DURATION + Exponential(MEAN_DURATION);
		/* hack to curtail the upper end of distribution */
		//		if( dur > 12000.0 )
		//		  dur /= 2.0;
	      }

	    /* Determine if first user is at home or on campus */
	    uni = Uniform01();
	    if( uni < HOME_PROB )
	      source = 'X';
	    else source = 'P';

	    /* determine if second user is at home or on campus */
	    uni = Uniform01();
	    if( uni < HOME_PROB )
	      dest = 'Y';
	    else dest = 'Q';

	    /* print out the call info */
#ifdef DEBUG
	    /* full debug info, including meeting ID and participants */
	    printf("%8.6f %d %d %d %c %c %8.6f\n",
		   timer, i, conns, size, source, dest, dur);
#else	
	    /* short version, ready to be used for workload.dat file */
	    printf("%8.6f %c %c %8.6f\n", timer, source, dest, dur);
#endif
#ifdef MEETINGS_VIEW
	    /* use lower case "m" to denote small P2P meetings */
	    printf("%8.6f %d %d %d %c %c %8.6f meeting\n",
		   timer, i, conns, size, source, dest, dur);
#endif
	    conns++;
	  }
	else /* not a P2P meeting */
	  {
	    /* Generate duration for meeting */
	    uni = Uniform01();
	    if( (uni < CLASS_PROB) && (timer > 27000.0) )
	      {
		/* cluster tightly around one hour duration */
		dur = CLASS_DURATION + Normal(0,600.0);
		while( dur < MEAN_DURATION )
		  dur = CLASS_DURATION + Normal(0, 600.0);
		
		/* add slight bias (longer durations) for large classes or evening ones (4-7pm) */
		if( (size > 30) || ((timer > 57600.0) && (timer < 68400.0)) )
		  {
		    dur = CLASS_DURATION + Exponential(CLASS_DURATION);;
		    /* small hack: reject shorter ones to boost variance a bit */
		    //		    while( (dur < CLASS_DURATION) || (dur > MAX_DURATION) )
		    //		      dur = Exponential(2.0*CLASS_DURATION);
		  }
	      }
	    else /* unconstrained duration, since not a lecture class */
	      {
		/* add slight bias (longer durations) for early morning and late evening */
		if( timer < 27000.0 )
		    dur = Exponential(2.0*MEAN_DURATION);
		else if( timer > 57600.0 )
		  dur = Uniform01()*4.0*MEAN_DURATION + Exponential(MEAN_DURATION);
		else dur = Exponential(1.5*MEAN_DURATION);
		/* small hack: reject shorter ones to boost variance a bit */
		//		while( dur < MEAN_DURATION)
		//		  dur = Exponential(1.5*MEAN_DURATION);
	      }
	    endtime = timer + dur;

	    /* determine location of Zoom server */
	    uni = Uniform01();
	    if( uni < VANCOUVER_PROB )
	      dest = 'V';
	    else if( uni < VANCOUVER_PROB + TORONTO_PROB )
	      dest = 'T';
	    else dest = 'O';

	    /* generate the conns for all the users in this Zoom meeting */
	    flag = 0;
	    while( size > 0 )
	      {
		/* determine if user is at home or on campus */
		uni = Uniform01();
		if( uni < HOME_PROB )
		  source = 'H';
		else
		  {
		    /* determine if user is on WiFi or Ethernet */
		    uni = Uniform01();
		    if( uni < WIFI_PROB )
		      source = 'W';
		    else source = 'E';
		  }

		/* adjust start time and duration for non-host participants */
		if( flag == 0 )
		  {
#ifdef MEETINGS_VIEW
		    /* use upper case "M" to denote larger non-P2P meetings */
		    printf("%8.6f %d %d %d %c %c %8.6f Meeting\n",
			   timer, i, conns, size, source, dest, dur);
#endif
		    starttime = timer;
		  }
		else starttime = timer + Normal(0.0, 60.0);
		flag++;

		/* print out the Zoom call info */
#ifdef DEBUG
		/* full debug info, including meeting ID and conn ID */
		printf("%8.6f %d %d %d %c %c %8.6f\n",
		       starttime, i, conns, size, source, dest, endtime - starttime);
#else	
		/* short version, ready to be used for workload.dat file */
		printf("%8.6f %c %c %8.6f\n", starttime, source, dest, endtime - starttime);
#endif
		conns++;
		size--;
#ifdef MEETINGS_ONLY
		size = 0;  /* hack to exit loop (i.e., meetings only, not conns) */
#endif
	      }
	  }

	/* Generate inter-arrival time from an exponential distribution */
	/* with time-varying rate based on time of day. */
	if( timer < 27000.0 )
	  iat = Exponential(MEAN_IAT*5.0); /* very light load before 7:30am */
	else if( timer > 75600.0 )
		  iat = Exponential(MEAN_IAT*4.0); /* light load after 9:00pm */
	else if( timer > 59400.0 )
		  iat = Exponential(MEAN_IAT/1.6); /* slightly reduced load after 4:30pm */
	else if( (int)timer%3600 > 3200 )
		  iat = Exponential(SHORT_IAT);  /* spikes at class start times 8am to 4pm */
	else if( (int)timer%1800 > 1500 )
		  iat = Exponential(SHORT_IAT*3.0);  /* smaller spikes at half-hour intervals */
	else iat = Exponential(MEAN_IAT/2.0); /* higher load from 8am to 5pm or so*/
	timer += iat;

      } /* end of for loop */

  } /* end of main */
